import inspect

from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import FeatureRequest, FeatureResultSet


class FeaturesEndpoint(UserBaseEndpoint):

    BASE_FEATURE_CRITERIA = {"stats": ["sum", "count"], "phq_rank": None}

    @classmethod
    def mutate_bool_to_default_for_type(cls, user_request_spec):
        attributes = inspect.getmembers(FeatureRequest, lambda a: not (inspect.isroutine(a)))
        fields_to_mutate = [
            a[0] for a in attributes if a[0].startswith("phq_attendance_") or a[0].startswith("phq_viewership_sports")
        ]
        for key, val in user_request_spec.items():
            if key in fields_to_mutate and isinstance(val, bool):
                user_request_spec[key] = cls.BASE_FEATURE_CRITERIA

    @accepts(FeatureRequest, query_string=False)
    @returns(FeatureResultSet)
    def obtain_features(self, **request):
        return self.client.post(self.build_url("v1", "features"), json=request)



from predicthq.endpoints.base import BaseEndpoint
from predicthq.endpoints.decorators import returns
from .schemas import Account


class AccountsEndpoint(BaseEndpoint):
    @returns(Account)
    def self(self):
        return self.client.get(self.build_url("v1", "accounts/self"))



from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import BroadcastResultSet, SearchParams


class BroadcastsEndpoint(UserBaseEndpoint):
    @accepts(SearchParams)
    @returns(BroadcastResultSet)
    def search(self, **params):
        return self.client.get(self.build_url("v1", "broadcasts"), params=params)



import functools
from collections import defaultdict

from predicthq.endpoints.schemas import ResultSet, Model, SchematicsDataError
from predicthq.exceptions import ValidationError


def _to_url_params(data, glue=".", separator=","):
    """
    Converts data dictionary to url parameters
    """
    params = {}
    for key, value in data.items():
        if isinstance(value, bool):
            params[key] = 1 if value else 0
        elif isinstance(value, list):
            params[key] = separator.join(map(str, value))
        elif isinstance(value, dict):
            for subkey, subvalue in value.items():
                if isinstance(subvalue, list):
                    params[glue.join((key, subkey))] = separator.join(map(str, subvalue))
                else:
                    params[glue.join((key, subkey))] = subvalue
        else:
            params[key] = value
    return params


def _process_kwargs(kwargs, separator="__"):
    data = defaultdict(dict)
    for key, value in kwargs.items():
        if separator in key:
            k, subk = key.split(separator)
            data[k][subk] = value
        else:
            data[key] = value
    return data


def accepts(schema_class, query_string=True, role=None):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):

            schema = getattr(endpoint.Meta, f.__name__, {}).get("accepts", schema_class)

            if not kwargs:  # accept instance of schema_class
                new_args = tuple(a for a in args if not isinstance(a, (schema, dict)))
                if args != new_args:
                    instance = next(a for a in args if isinstance(a, (schema, dict)))
                    if isinstance(instance, dict):
                        kwargs = instance
                    else:
                        kwargs = instance.to_dict()
                    args = new_args

            try:
                data = _process_kwargs(kwargs)
                if hasattr(endpoint, "mutate_bool_to_default_for_type"):
                    endpoint.mutate_bool_to_default_for_type(data)
                model = schema()
                model.import_data(data, strict=True, partial=False)
                model.validate()
            except SchematicsDataError as e:
                raise ValidationError(e.messages)

            if query_string:
                params = _to_url_params(model.to_primitive(role=role))
            else:
                params = model.to_primitive(role=role)

            return f(endpoint, *args, **params)

        return wrapper

    return decorator


def returns(schema_class):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):

            schema = getattr(endpoint.Meta, f.__name__, {}).get("returns", schema_class)

            data = f(endpoint, *args, **kwargs)
            try:
                model = schema()
                model._endpoint = endpoint

                # if schema class is a ResultSet, tell it how to load more results
                if issubclass(schema_class, ResultSet):
                    model._more = functools.partial(wrapper, endpoint)

                    # if results are of type Model, make sure to set the endpoint on each item
                    if (
                        data is not None
                        and "results" in data
                        and hasattr(model._fields["results"], "model_class")
                        and issubclass(model._fields["results"].model_class, Model)
                    ):

                        def initialize_result_type(item_data):
                            item = model._fields["results"].model_class(item_data, strict=False)
                            item._endpoint = endpoint
                            return item

                        # Use generator so results are not iterated over more than necessary
                        data["results"] = (initialize_result_type(item_data) for item_data in data["results"])

                model.import_data(data, strict=False)
                model.validate()
            except SchematicsDataError as e:
                raise ValidationError(e.messages)
            return model

        return wrapper

    return decorator



from predicthq.endpoints.base import BaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import SearchParams, PlaceResultSet


class PlacesEndpoint(BaseEndpoint):
    @accepts(SearchParams)
    @returns(PlaceResultSet)
    def search(self, **params):
        return self.client.get(self.build_url("v1", "places"), params=params)



from predicthq.endpoints.schemas import Model, StringType, DateTimeType, ModelType


class Industry(Model):

    id = StringType()
    name = StringType()


class Account(Model):

    id = StringType()
    name = StringType()
    description = StringType()
    industry = ModelType(Industry)
    created_at = DateTimeType()
    updated_at = DateTimeType()



from predicthq.config import config
from predicthq.endpoints.schemas import Model, StringType, StringListType, IntType


class GetTokenParams(Model):
    class Options:
        serialize_when_none = False

    client_id = StringType(default=lambda: config.OAUTH2_CLIENT_ID, required=True)
    client_secret = StringType(default=lambda: config.OAUTH2_CLIENT_SECRET, required=True)
    scope = StringListType(StringType, default=lambda: config.OAUTH2_SCOPE, separator=" ")
    grant_type = StringType(choices=("client_credentials",), default="client_credentials", required=True)


class RevokeTokenParams(Model):
    class Options:
        serialize_when_none = False

    client_id = StringType(default=lambda: config.OAUTH2_CLIENT_ID, required=True)
    client_secret = StringType(default=lambda: config.OAUTH2_CLIENT_SECRET, required=True)
    token = StringType(required=True)
    token_type_hint = StringType(choices=("access_token", "refresh_token"), default="access_token", required=True)


class AccessToken(Model):

    access_token = StringType()
    token_type = StringType()
    scope = StringListType(StringType, separator=" ")
    refresh_token = StringType(serialize_when_none=False)
    expires_in = IntType(serialize_when_none=False)



from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import (
    SearchParams,
    EventResultSet,
    CountResultSet,
    CalendarParams,
    CalendarResultSet,
    ImpactParams,
    ImpactResultSet,
)


class EventsEndpoint(UserBaseEndpoint):
    @accepts(SearchParams)
    @returns(EventResultSet)
    def search(self, **params):
        return self.client.get(self.build_url("v1", "events"), params=params)

    @accepts(SearchParams)
    @returns(CountResultSet)
    def count(self, **params):
        return self.client.get(self.build_url("v1", "events/count"), params=params)

    @accepts(CalendarParams)
    @returns(CalendarResultSet)
    def calendar(self, **params):
        return self.client.get(self.build_url("v1", "events/calendar"), params=params)

    @accepts(ImpactParams)
    @returns(ImpactResultSet)
    def impact(self, **params):
        return self.client.get(self.build_url("v1", "events/impact"), params=params)



from predicthq.endpoints.schemas import (
    PaginatedMixin,
    SortableMixin,
    BooleanType,
    DateTimeRange,
    DateTimeType,
    FloatType,
    IntRange,
    IntType,
    ListType,
    Model,
    ModelType,
    ResultSet,
    ResultType,
    StringType,
)


class BroadcastEventParams(Model):
    class Options:
        serialize_when_none = False

    event_id = ListType(StringType)
    category = ListType(StringType)
    label = ListType(StringType)


class BroadcastLocationParams(Model):
    class Options:
        serialize_when_none = False

    origin = origin = StringType(regex=r"(-?\d+(\.\d+)?),(-?\d+(\.\d+)?)")
    place_id = ListType(StringType)


class SearchParams(PaginatedMixin, SortableMixin, Model):
    class Options:
        serialize_when_none = False

    broadcast_id = ListType(StringType)
    location = ModelType(BroadcastLocationParams)
    phq_viewership = ModelType(IntRange)
    start = ModelType(DateTimeRange)
    updated = ModelType(DateTimeRange)
    first_seen = ModelType(DateTimeRange)
    record_status = ListType(StringType(choices=("active", "deleted"), default="active"))
    broadcast_status = ListType(StringType(choices=("scheduled", "predicted", "cancelled")))
    event = ModelType(BroadcastEventParams)


class GeoPoint(Model):

    lat = FloatType()
    lon = FloatType()


class BroadcastEventEntities(Model):
    class Options:
        serialize_when_none = False

    entity_id = StringType()
    type = StringType()
    name = StringType()
    formatted_address = StringType()


class BroadcastEventLocation(Model):
    class Options:
        serialize_when_none = False

    geopoint = ModelType(GeoPoint)
    place_hierarchies = ListType(ListType(StringType))
    country = StringType()


class BroadcastEventDates(Model):
    class Options:
        serialize_when_none = False

    start = DateTimeType()
    end = DateTimeType()
    start_local = DateTimeType()
    end_local = DateTimeType()
    timezone = StringType()

    # predicted_end_local is a paid feature.
    # It will only show up in your response body if you
    # have subscribed to it.
    predicted_end_local = DateTimeType()


class BroadcastEvent(Model):
    class Options:
        serialize_when_none = False

    event_id = StringType()
    title = StringType()
    category = StringType()
    labels = ListType(StringType)
    dates = ModelType(BroadcastEventDates)
    location = ModelType(BroadcastEventLocation)
    entities = ListType(ModelType(BroadcastEventEntities))

    # The following fields are paid features.
    # They will only show up in your response body if you
    # have subscribed to them.
    phq_attendance = IntType()
    phq_rank = IntType()
    local_rank = IntType()
    aviation_rank = IntType()


class Place(Model):
    class Options:
        serialize_when_none = False

    place_id = StringType()
    type = StringType()
    name = StringType()
    county = StringType()
    region = StringType()
    country = StringType()


class BroadcastLocation(Model):
    class Options:
        serialize_when_none = False

    geopoint = ModelType(GeoPoint)
    place_hierarchies = ListType(ListType(StringType))
    places = ListType(ModelType(Place))
    country = StringType()


class BroadcastDates(Model):
    class Options:
        serialize_when_none = False

    start = DateTimeType()
    start_local = DateTimeType()
    timezone = StringType()


class Broadcast(Model):
    class Options:
        serialize_when_none = False

    broadcast_id = StringType()
    updated = DateTimeType()
    first_seen = DateTimeType()
    dates = ModelType(BroadcastDates)
    location = ModelType(BroadcastLocation)
    phq_viewership = IntType()
    record_status = StringType()
    broadcast_status = StringType()
    event = ModelType(BroadcastEvent)


class BroadcastResultSet(ResultSet):

    overflow = BooleanType()
    results = ResultType(Broadcast)



from predicthq.endpoints.base import BaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import AccessToken, GetTokenParams, RevokeTokenParams


class OAuth2Endpoint(BaseEndpoint):
    @accepts(GetTokenParams)
    @returns(AccessToken)
    def get_token(self, client_id, client_secret, scope, grant_type, **kwargs):
        data = {
            "grant_type": grant_type,
            "scope": scope,
        }
        data.update(kwargs)
        return self.client.post("/oauth2/token/", auth=(client_id, client_secret), data=data)

    @accepts(RevokeTokenParams)
    def revoke_token(self, client_id, client_secret, token, token_type_hint):
        data = {
            "token_type_hint": token_type_hint,
            "token": token,
        }
        self.client.post("/oauth2/revoke/", auth=(client_id, client_secret), data=data)



import configparser
import os

from predicthq.exceptions import ConfigError


CONFIG_LOCATIONS = (
    "/etc/predicthq.conf",
    os.path.join(os.path.expanduser("~"), ".predicthq", "predicthq.conf"),
    os.path.join(os.getcwd(), "predicthq.conf"),
)


class Config(object):

    _config_sections = (
        "endpoint",
        "oauth2",
        "logging",
    )

    _defaults = {
        "ENDPOINT_URL": "https://api.predicthq.com",
        "OAUTH2_CLIENT_ID": None,
        "OAUTH2_CLIENT_SECRET": None,
        "OAUTH2_SCOPE": None,
        "OAUTH2_ACCESS_TOKEN": None,
        "LOGGING_LOG_LEVEL": "WARNING",
    }

    _config = {}

    def __init__(self, *config_locations):
        self.load_defaults()
        self.load_defaults_from_locations(config_locations)
        self.load_defaults_from_environment()

    def load_defaults(self):
        for key, value in self._defaults.items():
            self._config[key] = value

    def load_defaults_from_locations(self, config_locations):
        cp = configparser.ConfigParser()
        for file_location in config_locations:
            if os.path.isfile(file_location):
                with open(file_location) as f:
                    cp.read_file(f)
                for section in self._config_sections:
                    try:
                        for key, value in cp.items(section):
                            self._config["{0}_{1}".format(section.upper(), key.upper())] = value
                    except configparser.NoSectionError:  # pragma: nocover
                        pass

    def load_defaults_from_environment(self):
        for key in self._defaults:
            self._config[key] = os.getenv("PREDICTHQ_{}".format(key), self._config[key])

    def __getattr__(self, item):
        try:
            return self._config[item]
        except KeyError:
            raise ConfigError("{} is not a valid config key".format(item))

    def __setattr__(self, item, value):
        if item not in self._config:
            raise ConfigError("{} is not a valid config key".format(item))
        self._config[item] = value


config = Config(*CONFIG_LOCATIONS)



from predicthq.endpoints.schemas import (
    BooleanType,
    DateType,
    DateRange,
    DictType,
    FloatType,
    IntRange,
    IntType,
    ListType,
    Model,
    ModelType,
    ResultSet,
    ResultType,
    StringType,
)


class FeatureCriteria(Model):
    class Options:
        serialize_when_none = True

    stats = ListType(StringType(choices=("avg", "count", "max", "median", "min", "sum", "std_dev"), default="count"))
    phq_rank = ModelType(IntRange)


class FeatureGeoPoint(Model):
    lat = FloatType(required=True, min_value=-90.0, max_value=90.0)
    lon = FloatType(required=True, min_value=-180.0, max_value=180.0)
    radius = StringType(
        regex=r"^\d+(k?m|mi)$",
        required=True,
        messages={
            "regex": "Radius needs to define a number and unit (m, km, mi) eg. 100km",
        },
    )


class FeatureLocation(Model):
    class Options:
        serialize_when_none = False

    geo = ModelType(FeatureGeoPoint)
    place_id = ListType(StringType)


class HourOfDayRange(Model):
    class Options:
        serialize_when_none = False

    gt = IntType(min_value=0, max_value=23)
    gte = IntType(min_value=0, max_value=23)
    lt = IntType(min_value=0, max_value=23)
    lte = IntType(min_value=0, max_value=23)


class FeatureRequest(Model):
    class Options:
        serialize_when_none = False

    active = ModelType(DateRange, required=True)
    hour_of_day_active = ModelType(HourOfDayRange)
    hour_of_day_start = ModelType(HourOfDayRange)
    hour_of_day_end = ModelType(HourOfDayRange)
    location = ModelType(FeatureLocation, required=True)
    # Attendance based feature criteria
    phq_attendance_academic_graduation = ModelType(FeatureCriteria)
    phq_attendance_academic_social = ModelType(FeatureCriteria)
    phq_attendance_community = ModelType(FeatureCriteria)
    phq_attendance_concerts = ModelType(FeatureCriteria)
    phq_attendance_conferences = ModelType(FeatureCriteria)
    phq_attendance_expos = ModelType(FeatureCriteria)
    phq_attendance_festivals = ModelType(FeatureCriteria)
    phq_attendance_performing_arts = ModelType(FeatureCriteria)
    phq_attendance_sports = ModelType(FeatureCriteria)
    # Rank based feature criteria
    phq_rank_daylight_savings = BooleanType()
    phq_rank_health_warnings = BooleanType()
    phq_rank_observances = BooleanType()
    phq_rank_public_holidays = BooleanType()
    phq_rank_school_holidays = BooleanType()
    phq_rank_politics = BooleanType()
    phq_rank_academic_session = BooleanType()
    phq_rank_academic_exam = BooleanType()
    phq_rank_academic_holiday = BooleanType()
    # Viewership based feature criteria
    phq_viewership_sports = ModelType(FeatureCriteria)
    phq_viewership_sports_american_football = ModelType(FeatureCriteria)
    phq_viewership_sports_american_football_ncaa_men = ModelType(FeatureCriteria)
    phq_viewership_sports_american_football_nfl = ModelType(FeatureCriteria)
    phq_viewership_sports_auto_racing = ModelType(FeatureCriteria)
    phq_viewership_sports_auto_racing_indy_car = ModelType(FeatureCriteria)
    phq_viewership_sports_auto_racing_nascar = ModelType(FeatureCriteria)
    phq_viewership_sports_baseball = ModelType(FeatureCriteria)
    phq_viewership_sports_baseball_mlb = ModelType(FeatureCriteria)
    phq_viewership_sports_baseball_ncaa_men = ModelType(FeatureCriteria)
    phq_viewership_sports_basketball = ModelType(FeatureCriteria)
    phq_viewership_sports_basketball_ncaa_women = ModelType(FeatureCriteria)
    phq_viewership_sports_basketball_ncaa_men = ModelType(FeatureCriteria)
    phq_viewership_sports_basketball_nba = ModelType(FeatureCriteria)
    phq_viewership_sports_boxing = ModelType(FeatureCriteria)
    phq_viewership_sports_golf = ModelType(FeatureCriteria)
    phq_viewership_sports_golf_masters = ModelType(FeatureCriteria)
    phq_viewership_sports_golf_pga_championship = ModelType(FeatureCriteria)
    phq_viewership_sports_golf_pga_tour = ModelType(FeatureCriteria)
    phq_viewership_sports_golf_us_open = ModelType(FeatureCriteria)
    phq_viewership_sports_horse_racing = ModelType(FeatureCriteria)
    phq_viewership_sports_horse_racing_belmont_stakes = ModelType(FeatureCriteria)
    phq_viewership_sports_horse_racing_kentucky_derby = ModelType(FeatureCriteria)
    phq_viewership_sports_horse_racing_preakness_stakes = ModelType(FeatureCriteria)
    phq_viewership_sports_ice_hockey = ModelType(FeatureCriteria)
    phq_viewership_sports_ice_hockey_nhl = ModelType(FeatureCriteria)
    phq_viewership_sports_mma = ModelType(FeatureCriteria)
    phq_viewership_sports_mma_ufc = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_concacaf_champions_league = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_concacaf_gold_cup = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_copa_america_men = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_fifa_world_cup_women = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_fifa_world_cup_men = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_mls = ModelType(FeatureCriteria)
    phq_viewership_sports_soccer_uefa_champions_league_men = ModelType(FeatureCriteria)
    phq_viewership_sports_softball = ModelType(FeatureCriteria)
    phq_viewership_sports_softball_ncaa_women = ModelType(FeatureCriteria)
    phq_viewership_sports_tennis = ModelType(FeatureCriteria)
    phq_viewership_sports_tennis_us_open = ModelType(FeatureCriteria)
    phq_viewership_sports_tennis_wimbledon = ModelType(FeatureCriteria)


class FeatureRankLevel(Model):
    class Options:
        serialize_when_none = False

    rank_levels = DictType(StringType)


class FeatureStats(Model):
    class Options:
        serialize_when_none = False

    avg = FloatType()
    count = IntType()
    max = FloatType()
    median = FloatType()
    min = FloatType()
    sum = FloatType()
    std_dev = FloatType()


class FeatureStat(Model):
    class Options:
        serialize_when_none = False

    stats = ModelType(FeatureStats)


class Feature(Model):
    class Options:
        serialize_when_none = False

    date = DateType()
    # Attendance based features
    phq_attendance_academic_graduation = ModelType(FeatureStat)
    phq_attendance_academic_social = ModelType(FeatureStat)
    phq_attendance_community = ModelType(FeatureStat)
    phq_attendance_concerts = ModelType(FeatureStat)
    phq_attendance_conferences = ModelType(FeatureStat)
    phq_attendance_expos = ModelType(FeatureStat)
    phq_attendance_festivals = ModelType(FeatureStat)
    phq_attendance_performing_arts = ModelType(FeatureStat)
    phq_attendance_sports = ModelType(FeatureStat)
    # Rank based features
    phq_rank_daylight_savings = ModelType(FeatureRankLevel)
    phq_rank_health_warnings = ModelType(FeatureRankLevel)
    phq_rank_observances = ModelType(FeatureRankLevel)
    phq_rank_public_holidays = ModelType(FeatureRankLevel)
    phq_rank_school_holidays = ModelType(FeatureRankLevel)
    phq_rank_politics = ModelType(FeatureRankLevel)
    phq_rank_academic_session = ModelType(FeatureRankLevel)
    phq_rank_academic_exam = ModelType(FeatureRankLevel)
    phq_rank_academic_holiday = ModelType(FeatureRankLevel)
    # Viewership based features
    phq_viewership_sports = ModelType(FeatureStat)
    phq_viewership_sports_american_football = ModelType(FeatureStat)
    phq_viewership_sports_american_football_ncaa_men = ModelType(FeatureStat)
    phq_viewership_sports_american_football_nfl = ModelType(FeatureStat)
    phq_viewership_sports_auto_racing = ModelType(FeatureStat)
    phq_viewership_sports_auto_racing_indy_car = ModelType(FeatureStat)
    phq_viewership_sports_auto_racing_nascar = ModelType(FeatureStat)
    phq_viewership_sports_baseball = ModelType(FeatureStat)
    phq_viewership_sports_baseball_mlb = ModelType(FeatureStat)
    phq_viewership_sports_baseball_ncaa_men = ModelType(FeatureStat)
    phq_viewership_sports_basketball = ModelType(FeatureStat)
    phq_viewership_sports_basketball_ncaa_women = ModelType(FeatureStat)
    phq_viewership_sports_basketball_ncaa_men = ModelType(FeatureStat)
    phq_viewership_sports_basketball_nba = ModelType(FeatureStat)
    phq_viewership_sports_boxing = ModelType(FeatureStat)
    phq_viewership_sports_golf = ModelType(FeatureStat)
    phq_viewership_sports_golf_masters = ModelType(FeatureStat)
    phq_viewership_sports_golf_pga_championship = ModelType(FeatureStat)
    phq_viewership_sports_golf_pga_tour = ModelType(FeatureStat)
    phq_viewership_sports_golf_us_open = ModelType(FeatureStat)
    phq_viewership_sports_horse_racing = ModelType(FeatureStat)
    phq_viewership_sports_horse_racing_belmont_stakes = ModelType(FeatureStat)
    phq_viewership_sports_horse_racing_kentucky_derby = ModelType(FeatureStat)
    phq_viewership_sports_horse_racing_preakness_stakes = ModelType(FeatureStat)
    phq_viewership_sports_ice_hockey = ModelType(FeatureStat)
    phq_viewership_sports_ice_hockey_nhl = ModelType(FeatureStat)
    phq_viewership_sports_mma = ModelType(FeatureStat)
    phq_viewership_sports_mma_ufc = ModelType(FeatureStat)
    phq_viewership_sports_soccer = ModelType(FeatureStat)
    phq_viewership_sports_soccer_concacaf_champions_league = ModelType(FeatureStat)
    phq_viewership_sports_soccer_concacaf_gold_cup = ModelType(FeatureStat)
    phq_viewership_sports_soccer_copa_america_men = ModelType(FeatureStat)
    phq_viewership_sports_soccer_fifa_world_cup_women = ModelType(FeatureStat)
    phq_viewership_sports_soccer_fifa_world_cup_men = ModelType(FeatureStat)
    phq_viewership_sports_soccer_mls = ModelType(FeatureStat)
    phq_viewership_sports_soccer_uefa_champions_league_men = ModelType(FeatureStat)
    phq_viewership_sports_softball = ModelType(FeatureStat)
    phq_viewership_sports_softball_ncaa_women = ModelType(FeatureStat)
    phq_viewership_sports_tennis = ModelType(FeatureStat)
    phq_viewership_sports_tennis_us_open = ModelType(FeatureStat)
    phq_viewership_sports_tennis_wimbledon = ModelType(FeatureStat)


class FeatureResultSet(ResultSet):
    results = ResultType(Feature)



from predicthq.endpoints.schemas import (
    LimitMixin,
    Model,
    ResultSet,
    ListType,
    StringType,
    GeoJSONPointType,
    StringListType,
    StringModelType,
    Location,
    DateTimeType,
    ResultType,
    SchematicsValidationError,
)


class SearchParams(LimitMixin, Model):
    class Options:
        serialize_when_none = False

    q = StringType()
    id = ListType(StringType)
    location = StringListType(StringModelType(Location), separator="+")
    country = ListType(StringType)
    type = ListType(
        StringType(
            choices=(
                "planet",
                "continent",
                "country",
                "region",
                "county",
                "localadmin",
                "locality",
                "neighbourhood",
                "local",
                "metro",
                "major",
                "all",
            )
        )
    )

    def validate(self, *args, **kwargs):
        super(SearchParams, self).validate(*args, **kwargs)
        if not any((self.q, self.id, self.location, self.country)):
            raise SchematicsValidationError("Places search requires one of q, id, location or country")


class Place(Model):

    id = StringType()
    type = StringType()
    name = StringType()
    county = StringType()
    region = StringType()
    country = StringType()
    country_alpha2 = StringType()
    country_alpha3 = StringType()
    location = GeoJSONPointType()


class PlaceResultSet(ResultSet):

    results = ResultType(Place)



import re
from datetime import datetime, date, time as dt_time
from urllib.parse import parse_qsl, urlparse

import pytz
from dateutil.parser import parse as parse_date

from schematics.exceptions import (
    ValidationError as SchematicsValidationError,
    DataError as SchematicsDataError,
    ConversionError,
)
from schematics.models import Model
from schematics.transforms import Role
from schematics.types import (
    StringType,
    DateTimeType as SchematicsDateTimeType,
    IntType,
    FloatType,
    URLType,
    GeoPointType,
    BooleanType,
    DateType as SchematicsDateType,
)
from schematics.types.compound import ListType as SchematicsListType, ModelType, DictType, PolyModelType
from schematics.types.serializable import serializable


class DateTimeType(SchematicsDateTimeType):
    def to_native(self, value, context=None):
        if isinstance(value, datetime):
            return value
        return parse_date(value)


class DateTimeEndType(SchematicsDateTimeType):
    def to_native(self, value, context=None):
        if isinstance(value, datetime):
            return value
        return datetime.combine(parse_date(value), dt_time.max)


class DateType(SchematicsDateType):
    def to_native(self, value, context=None):
        if isinstance(value, date):
            return value
        return parse_date(value).date()


class StringModelType(ModelType):
    def _convert(self, value, context):
        if isinstance(value, str):
            return self.model_class._convert(value, context=context)
        else:
            return super(StringModelType, self)._convert(value, context=context)


class StringModel(Model):
    @property
    def import_format(self):  # pragma: nocover
        raise NotImplementedError()

    @property
    def export_format(self):  # pragma: nocover
        raise NotImplementedError()

    def convert(self, raw_data, **kw):
        if isinstance(raw_data, str):
            try:
                raw_data = re.match(self.import_format, str(raw_data)).groupdict()
            except AttributeError:
                raise SchematicsValidationError(
                    "'{}' is not a valid format for {}".format(raw_data, self.__class__.__name__)
                )
        return super(StringModel, self).convert(raw_data, **kw)

    def export(self, *args, **kwargs):
        return self.export_format.format(**super(StringModel, self).export(*args, **kwargs))


class StringListType(SchematicsListType):
    def __init__(self, field, separator=" ", **kwargs):
        self.separator = separator
        super(StringListType, self).__init__(field, **kwargs)

    def _coerce(self, value):
        if isinstance(value, str):
            return value.split(self.separator)
        elif isinstance(value, dict):
            return [value]
        else:
            return super(StringListType, self)._coerce(value)

    def _export(self, *args, **kwargs):
        data = super(StringListType, self)._export(*args, **kwargs)
        return self.separator.join(map(str, data))


class ListType(SchematicsListType):
    def _coerce(self, value):
        if not isinstance(value, (list, tuple)):
            return [value]
        else:
            return super(ListType, self)._coerce(value)


class GeoJSONPointType(GeoPointType):
    def _normalize(self, value):
        return tuple(reversed(super(GeoJSONPointType, self)._normalize(value)))


class Area(StringModel):

    import_format = r"(?P<radius>\d+(k?m|ft|mi))@(?P<latitude>-?\d+(\.\d+)?),(?P<longitude>-?\d+(\.\d+)?)"
    export_format = "{radius}@{latitude},{longitude}"

    radius = StringType(regex=r"\d+(k?m|ft|mi)", required=True)
    latitude = FloatType(required=True)
    longitude = FloatType(required=True)


class Location(StringModel):

    import_format = r"@(?P<latitude>-?\d+(\.\d+)?),(?P<longitude>-?\d+(\.\d+)?)"
    export_format = "@{latitude},{longitude}"

    latitude = FloatType(required=True)
    longitude = FloatType(required=True)


class Place(Model):
    class Options:
        serialize_when_none = False

    scope = ListType(StringType)
    exact = ListType(StringType)


class DateAround(Model):
    class Options:
        serialize_when_none = False

    origin = DateType(required=True)
    offset = StringType(regex=r"\d+d")
    scale = StringType(regex=r"\d+d")
    decay = FloatType()


class LocationAround(Model):
    class Options:
        serialize_when_none = False

    origin = StringType(regex=r"(-?\d+(\.\d+)?),(-?\d+(\.\d+)?)", required=True)
    offset = StringType(regex=r"\d+(\.\d+)?(cm|m|km|in|ft|mi)")
    scale = StringType(regex=r"\d+(\.\d+)?(cm|m|km|in|ft|mi)")
    decay = FloatType()


class DateTimeRange(Model):
    class Options:
        serialize_when_none = False

    gt = DateTimeType()
    gte = DateTimeType()
    lt = DateTimeEndType()
    lte = DateTimeEndType()
    tz = StringType(choices=pytz.all_timezones)


class DateRange(Model):
    class Options:
        serialize_when_none = False

    gt = DateType()
    gte = DateType()
    lt = DateType()
    lte = DateType()
    tz = StringType(choices=pytz.all_timezones)


class FloatRange(Model):
    class Options:
        serialize_when_none = False

    gt = FloatType()
    gte = FloatType()
    lt = FloatType()
    lte = FloatType()


class IntRange(Model):
    class Options:
        serialize_when_none = False

    gt = IntType()
    gte = IntType()
    lt = IntType()
    lte = IntType()


class BrandUnsafe(Model):
    class Options:
        serialize_when_none = False

    exclude = BooleanType()


class Entity(Model):
    class Options:
        serialize_when_none = False

    id = StringType()


class LimitMixin(Model):

    limit = IntType(min_value=1)


class OffsetMixin(Model):

    offset = IntType(min_value=0)


class PaginatedMixin(LimitMixin, OffsetMixin):

    pass


class SortableMixin(Model):

    sort = ListType(StringType())


class ResultSet(Model):

    count = IntType()
    previous = URLType()
    next = URLType()

    @property
    def results(self):  # pragma: nocover
        raise NotImplementedError()

    def _parse_params(self, url):
        return dict(parse_qsl(urlparse(url).query))

    def has_previous(self):
        return self.previous is not None

    def has_next(self):
        return self.next is not None

    def get_next(self):
        if not self.has_next() or not hasattr(self, "_more"):
            return
        params = self._parse_params(self.next)
        return self._more(**params)

    def get_previous(self):
        if not self.has_previous() or not hasattr(self, "_more"):
            return
        params = self._parse_params(self.previous)
        return self._more(**params)

    def iter_pages(self):
        next_page = self
        while next_page.has_next():
            next_page = next_page.get_next()
            yield next_page

    def iter_items(self):
        return iter(self.results)

    def iter_all(self):
        for item in self.iter_items():
            yield item
        for page in self.iter_pages():
            for item in page.iter_items():
                yield item

    def __iter__(self):
        return self.iter_items()


class ResultType(SchematicsListType):
    def __init__(self, field, *args, **kwargs):
        super(ResultType, self).__init__(ModelType(field), *args, **kwargs)



from schematics.common import NONEMPTY

from predicthq.endpoints.schemas import (
    PaginatedMixin,
    SortableMixin,
    Model,
    ResultSet,
    ListType,
    StringType,
    GeoJSONPointType,
    StringListType,
    StringModelType,
    Area,
    ModelType,
    IntRange,
    IntType,
    DateTimeRange,
    DateTimeType,
    FloatType,
    ResultType,
    DictType,
    DateType,
    Place,
    DateAround,
    LocationAround,
    BooleanType,
    BrandUnsafe,
    Entity,
    PolyModelType,
)


class SearchParams(PaginatedMixin, SortableMixin, Model):
    class Options:
        serialize_when_none = False

    active = ModelType(DateTimeRange)
    cancelled = ModelType(DateTimeRange)
    category = ListType(StringType)
    country = ListType(StringType)
    deleted_reason = ListType(StringType(choices=("cancelled", "duplicate", "invalid", "postponed")))
    end = ModelType(DateTimeRange)
    end_around = ModelType(DateAround)
    id = ListType(StringType)
    label = ListType(StringType)
    location_around = ModelType(LocationAround)
    place = ModelType(Place)
    postponed = ModelType(DateTimeRange)
    q = StringType()
    relevance = ListType(StringType)
    start = ModelType(DateTimeRange)
    start_around = ModelType(DateAround)
    state = ListType(StringType(choices=("active", "deleted"), default="active"))
    updated = ModelType(DateTimeRange)
    within = StringListType(StringModelType(Area), separator="+")

    # The below parameters are only available if they are enabled in your plan.
    # If you are not subscribed to a feature, using the parameter will have no
    # effect on your search results.
    aviation_rank = ModelType(IntRange)
    aviation_rank_level = ListType(IntType(min_value=1, max_value=5))
    brand_unsafe = ModelType(BrandUnsafe)
    entity = ModelType(Entity)
    local_rank = ModelType(IntRange)
    local_rank_level = ListType(IntType(min_value=1, max_value=5))
    phq_attendance = ModelType(IntRange)
    predicted_end = ModelType(DateTimeRange)
    rank = ModelType(IntRange)
    rank_level = ListType(IntType(min_value=1, max_value=5))


class Entities(Model):
    class Options:
        serialize_when_none = True

    entity_id = StringType()
    name = StringType()
    type = StringType()
    formatted_address = StringType()


class Point(Model):

    type = StringType()
    coordinates = ListType(FloatType())

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("type") in ["Point"]


class MultiPoint(Model):

    type = StringType()
    coordinates = ListType(ListType(FloatType()))

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("type") in ["MultiPoint", "LineString"]


class Polygon(Model):

    type = StringType()
    coordinates = ListType(ListType(ListType(FloatType())))

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("type") in ["MultiLineString", "Polygon"]


class MultiPolygon(Model):

    type = StringType()
    coordinates = ListType(ListType(ListType(ListType(FloatType()))))

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("type") in ["MultiPolygon"]


class Geo(Model):

    geometry = PolyModelType(model_spec=[Point, MultiPoint, Polygon, MultiPolygon])


class ParentEvent(Model):

    parent_event_id = StringType()


class Event(Model):
    class Options:
        serialize_when_none = True

    cancelled = DateTimeType()
    category = StringType()
    country = StringType()
    deleted_reason = StringType()
    description = StringType()
    duplicate_of_id = StringType()
    duration = IntType()
    end = DateTimeType()
    first_seen = DateTimeType()
    geo = ModelType(Geo)
    id = StringType()
    labels = ListType(StringType())
    location = GeoJSONPointType()
    parent_event = ModelType(ParentEvent)
    place_hierarchies = ListType(ListType(StringType()))
    postponed = DateTimeType()
    relevance = FloatType()
    scope = StringType()
    start = DateTimeType()
    state = StringType()
    timezone = StringType()
    title = StringType()
    updated = DateTimeType()

    # The below fields are only available if they are enabled in your plan.
    aviation_rank = IntType()  # Aviation Rank add-on
    brand_safe = BooleanType()
    entities = ListType(ModelType(Entities))  # Venues and addresses add-on
    local_rank = IntType()  # Local Rank add-on
    phq_attendance = IntType()  # PHQ Attendance add-on
    predicted_end = DateTimeType()
    private = BooleanType()  # Loop add-on
    rank = IntType()  # PHQ Rank add-on


class EventResultSet(ResultSet):

    overflow = BooleanType()

    results = ResultType(Event)


class CountResultSet(Model):

    count = IntType()
    top_rank = FloatType()
    rank_levels = DictType(IntType)
    categories = DictType(IntType)
    labels = DictType(IntType)


class TopEventsSearchParams(SortableMixin, Model):

    limit = IntType(min_value=0, max_value=10)


class CalendarParams(SearchParams):

    top_events = ModelType(TopEventsSearchParams)


class CalendarDay(Model):

    date = DateType()
    count = IntType()
    top_rank = FloatType()
    rank_levels = DictType(IntType)
    categories = DictType(IntType)
    labels = DictType(IntType)
    top_events = ModelType(EventResultSet)


class CalendarResultSet(ResultSet):

    results = ResultType(CalendarDay)


class ImpactParams(SearchParams):

    top_events = ModelType(TopEventsSearchParams)
    impact_rank = StringType(choices=("rank", "aviation_rank"))


class ImpactDay(Model):

    date = DateType()
    count = IntType()
    impact = IntType()

    rank_levels = DictType(IntType, export_level=NONEMPTY)
    rank_levels_impact = DictType(IntType, export_level=NONEMPTY)

    aviation_rank_levels = DictType(IntType, export_level=NONEMPTY)
    aviation_rank_levels_impact = DictType(IntType, export_level=NONEMPTY)

    categories = DictType(IntType)
    categories_impact = DictType(IntType)


class ImpactResultSet(ResultSet):

    results = ResultType(ImpactDay)
